import pytest
from hypothesis import given
import hypothesis.strategies as st
import hypothesis.extra.numpy as hypnp
import hypothesis as hyp
import cvxpy as cp

import numpy as np
#import soundfieldcontrol.szc.sinr_opt as sinropt
#import soundfieldcontrol.szc.sinr_opt_time as sinrt

import aspcol.soundfieldcontrol as sfc
import aspcol.unittests.matrix_strategies as mst


SEED = 5

NUM_ZONES = st.shared(st.integers(min_value=2, max_value=4))
NUM_LS = st.shared(st.integers(min_value=1, max_value=3))
FILT_LEN = st.shared(st.integers(min_value=1, max_value=3))

@st.composite
def SPATIAL_COV_TD(draw):
    num_zones = draw(NUM_ZONES)
    num_ls = draw(NUM_LS)
    filt_len = draw(FILT_LEN)
    dim = num_ls * filt_len
    cov_mat = np.stack([np.stack([mst.psd_real(draw, dim, dim) for _ in range(num_zones)], axis=0) for _ in range(num_zones)], axis=0)
    return cov_mat

@st.composite
def SINR_TARGETS(draw):
    num_zones = draw(NUM_ZONES)
    targets = draw(hypnp.arrays(dtype=float, shape=num_zones, elements=st.floats(min_value=1e-5, max_value = 100, allow_infinity=False, allow_nan=False)))
    return targets

@st.composite
def AUDIO_COV(draw):
    """
    should really be toeplitz psd matrix
    """
    num_zones = draw(NUM_ZONES)
    num_ls = draw(NUM_LS)
    filt_len = draw(FILT_LEN)
    dim = num_ls * filt_len

    regularization = draw(st.floats(min_value = 0, max_value=1))
    cov_mat = np.stack([mst.psd_real(draw, dim) + regularization * np.eye(dim) for _ in range(num_zones)], axis=0)
    return cov_mat

#def RANDOM_BEAMFORMER(draw):

# SPATIAL_COV_TD = hypnp.arrays(dtype=dtype_tuple[0], 
#                                 shape=st.tuples(st.integers(min_value=1, max_value=1), outer_dim_l, inner_dim),
#                                 elements=st.complex_numbers(allow_infinity=False, allow_nan=False))
#outer_dim_l = st.integers(min_value=1, max_value=5)
#outer_dim_r = st.integers(min_value=1, max_value=5)
#dtype_st = st.shared(st.sampled_from((float, complex)))


# dtype_tuple = st.shared(st.sampled_from((
#                         (float, st.floats(allow_infinity=False, allow_nan=False)), 
#                         (complex, st.complex_numbers(allow_infinity=False, allow_nan=False))
#                             )))

# mat_left_real = hypnp.arrays(dtype=dtype_tuple[0], 
#                                 shape=st.tuples(st.integers(min_value=1, max_value=1), outer_dim_l, inner_dim),
#                                 elements=st.complex_numbers(allow_infinity=False, allow_nan=False))
# mat_left_complex = hypnp.arrays(dtype=complex, 
#                                 shape=st.tuples(st.integers(min_value=1, max_value=1), outer_dim_l, inner_dim),
#                                 elements=st.complex_numbers(allow_infinity=False, allow_nan=False))
# matl = st.sampled_from(mat_left_real, mat_left_complex)

# mat_right_real = hypnp.arrays(dtype=dtype_st, shape=st.tuples(st.integers(min_value=1, max_value=1), inner_dim, outer_dim_r))



def get_random_spatial_cov(num_zones, num_ls):
    rng = np.random.default_rng()
    rank = rng.integers(1, num_ls)
    R = np.zeros((num_zones, num_ls, num_ls), dtype=complex)
    for k in range(num_zones):
        for r in range(rank):
            v = rng.normal(size=(num_ls, 1)) + 1j*rng.normal(size=(num_ls, 1))
            R[k,:,:] += v @ v.conj().T

    assert np.allclose(R, np.moveaxis(R.conj(), 1,2))
    return R

def get_random_spatial_cov_full_rank(num_freq, num_ls):
    rng = np.random.default_rng()
    rank = num_ls + 4
    R = np.zeros((num_freq, num_ls, num_ls), dtype=complex)
    for k in range(num_freq):
        for r in range(rank):
            v = rng.normal(size=(num_ls, 1)) + 1j*rng.normal(size=(num_ls, 1))
            R[k,:,:] += v @ v.conj().T

    assert np.allclose(R, np.moveaxis(R.conj(), 1,2))
    return R


def get_random_beamformer(num_zones, num_ls):
    rng = np.random.default_rng()
    w = rng.normal(scale=rng.uniform(high=10),size=(num_zones, num_ls))
    return w

def get_random_beamformer_normalized(num_zones, num_ls):
    rng = np.random.default_rng()
    w = rng.normal(scale=rng.uniform(high=10),size=(num_zones, num_ls)) + 1j*rng.normal(scale=rng.uniform(high=5),size=(num_zones, num_ls))
    w /= np.sqrt(np.sum(np.abs(w)**2, axis=-1, keepdims=True))
    return w

def get_random_noise_pow(num_zones):
    rng = np.random.default_rng()
    return rng.uniform(low=1e-3, high=10, size=num_zones)

def get_random_sinr_targets(num_zones):
    rng = np.random.default_rng()
    exponent = rng.uniform(-3, 1, size=num_zones)
    return 10**(exponent)
    
def get_random_pow_vec(num_zones):
    rng = np.random.default_rng()
    exponent = rng.uniform(-3, 1, size=num_zones)
    return 10**(exponent)


def get_random_spatial_cov_td(num_zones, bf_len):
    rng = np.random.default_rng()
    #rank = rng.integers(1, num_ls)
    R = np.zeros((num_zones, num_zones, bf_len, bf_len), dtype=float)
    for k in range(num_zones):
        for i in range(num_zones):
            for r in range(bf_len+5):
                v = rng.normal(size=(bf_len, 1))
                R[k,i,:,:] += v @ v.T

    assert np.allclose(R, np.moveaxis(R, 2,3))
    return R

def get_random_audio_cov(num_zones, bf_len):
    rng = np.random.default_rng()
    reg_param = rng.uniform(low=1e-5, high=1)
    R = np.zeros((num_zones, bf_len, bf_len), dtype=float)
    for k in range(num_zones):
        for r in range(bf_len):
            v = rng.normal(size=(bf_len, 1))
            R[k,:,:] += v @ v.T
    R += reg_param * np.eye(bf_len)[None,:,:]

    assert np.allclose(R, np.moveaxis(R, 1,2))
    return R


def get_random_beamformer_normalized_td(num_zones, bf_len):
    rng = np.random.default_rng()
    w = rng.normal(scale=rng.uniform(high=10),size=(num_zones, bf_len))
    w /= np.sqrt(np.sum(np.abs(w)**2, axis=-1, keepdims=True))
    return w

# @hyp.settings(deadline=None)
# @given(
#     st.integers(min_value=2, max_value=5),
#     st.integers(min_value=2, max_value=5),
# )
# def test_sinr_uplink_specific_and_link_gain_formulation_equal(num_zones, num_ls):
#     w = get_random_beamformer_normalized(num_zones, num_ls)
#     R = get_random_spatial_cov(num_zones, num_ls)
#     noise_pow = np.ones_like(get_random_noise_pow(num_zones))
#     p = get_random_pow_vec(num_zones)
#     #w = sinropt.apply_power_vec(w_norm, p)
#     #sinr_targets = get_random_sinr_targets(num_zones)

#     sinr_val = sfc._sinr_uplink(p, w, R, noise_pow)
#     #sinr_val1 = sinropt._sinr_uplink_comparison(w_norm, R, p, noise_pow)
#     sinr_val2 = sfc._sinr_uplink_comparison(p, w, R, noise_pow)
    
#     assert np.allclose(sinr_val, sinr_val2)



# @hyp.settings(deadline=None)
# @given(
#     st.integers(min_value=2, max_value=5),
#     st.integers(min_value=2, max_value=5),
# )
# def test_sinr_downlink_specific_and_link_gain_formulation_equal(num_zones, num_ls):
#     w = get_random_beamformer_normalized(num_zones, num_ls)
#     R = get_random_spatial_cov(num_zones, num_ls)
#     noise_pow = np.ones_like(get_random_noise_pow(num_zones))
#     p = get_random_pow_vec(num_zones)
#     #w = sinropt.apply_power_vec(w_norm, p)
#     #sinr_targets = get_random_sinr_targets(num_zones)

#     sinr_val = sfc._sinr_downlink(p, w, R, noise_pow)
#     sinr_val2 = sfc._sinr_downlink_comparison(sfc.apply_power_vec(w, p), R, noise_pow)
    
#     assert np.allclose(sinr_val, sinr_val2)


def sinr_constrained_weighted_pow_min_downlink(R_values, avg_noise_pow, sinr_targets, audio_cov_values):
    """
    Solves the SINR constrained power minimization problem for transmit beamforming
        via semidefinite relaxation. Returns the optimal matrix, not the optimal 
        beamformer vector. Use any of the select_solution functions for that. 

    R_values is the spatial covariance matrices R_{ki}, where k is the zone index of 
        the microphones and i is the zone index of the audio signal
        The array should have has shape (num_zones, num_zones, bf_len, bf_len)
        where bf_len = num_ls*ir_len, and is indexed as R_{ki} = R[k,i,:,:]

    avg_noise_pow is positive real values of shape (num_zones)
    sinr_targets is positive real values of shape (num_zones)

    returns opt_mat which has shape (num_zones, bf_len, bf_len)
    """
    num_zones = R_values.shape[1]
    mat_size = R_values.shape[2]
    assert R_values.shape == (num_zones, num_zones, mat_size, mat_size)
    assert len(avg_noise_pow) == num_zones
    assert len(sinr_targets) == num_zones

    opt_mat = np.zeros((num_zones, mat_size, mat_size), dtype=float)
    W = [cp.Variable((mat_size, mat_size), PSD=True) for _ in range(num_zones)]
    audio_cov = [cp.Parameter((mat_size, mat_size), PSD=True) for _ in range(num_zones)]
    R = [[cp.Parameter((mat_size, mat_size), PSD=True) for _ in range(num_zones)] for _ in range(num_zones)]
    sinr_target = [cp.Parameter(pos=True) for _ in range(num_zones)]
    noise_pow = [cp.Parameter(pos=True) for _ in range(num_zones)]
    constraints = []

    for k in range(num_zones):
        new_constr = cp.trace(R[k][k] @ W[k]) - sinr_target[k] * noise_pow[k]
        for i in range(num_zones):
            if k != i:
                new_constr -= sinr_target[k] * cp.trace(R[k][i] @ W[i])
        constraints.append(new_constr >= 0)
                    
    for k in range(num_zones):
        constraints.append(W[k] == W[k].T)

    obj = cp.Minimize(cp.sum([cp.trace(audio_cov_z @ W_z) for W_z, audio_cov_z in zip(W, audio_cov)]))
    prob = cp.Problem(obj, constraints)


    for k in range(num_zones):
        for i in range(num_zones):
            R[k][i].value = R_values[k,i,:,:]
        sinr_target[k].value = sinr_targets[k]
        noise_pow[k].value = avg_noise_pow[k]
        audio_cov[k].value = audio_cov_values[k]
    prob.solve(verbose=False, ignore_dpp=True, eps=1e-8, max_iters=10000)
    print(prob.status)
    for k in range(num_zones):
        opt_mat[k,:,:] = W[k].value
    return opt_mat




@hyp.settings(deadline=None)
@given(
    st.integers(min_value=2, max_value=5),
    st.integers(min_value=2, max_value=5),
    st.floats(min_value=1, max_value=1000),
)
def test_power_alloc_minmax_equal_to_max_pow(num_zones, num_ls, max_pow):
    w = get_random_beamformer_normalized(num_zones, num_ls)
    R = get_random_spatial_cov(num_zones, num_ls)
    noise_pow = get_random_noise_pow(num_zones)
    sinr_targets = get_random_sinr_targets(num_zones)

    p_dl, _ = sfc._power_alloc_minmax_downlink(w, R, noise_pow, sinr_targets, max_pow)
    assert np.allclose(np.sum(p_dl), max_pow)

    p_ul, _ = sfc._power_alloc_minmax_uplink(w, R, noise_pow, sinr_targets, max_pow)
    assert np.allclose(np.sum(p_ul), max_pow)



@hyp.settings(deadline=None)
@given(
    st.integers(min_value=2, max_value=5),
    st.integers(min_value=2, max_value=5),
)
def test_power_alloc_qos_zero_sinr_margin_downlink(num_zones, num_ls):
    num_test_attempts = 100
    feasible = False
    for i in range(num_test_attempts):
        w = get_random_beamformer_normalized(num_zones, num_ls)
        R = get_random_spatial_cov(num_zones, num_ls)
        noise_pow = get_random_noise_pow(num_zones)
        rng = np.random.default_rng(SEED+12)
        sinr_targets = rng.uniform(low=0.05, high=0.3, size=num_zones)
        if sfc._power_alloc_qos_is_feasible(sfc._link_gain_downlink(w, R), sinr_targets):
            feasible = True
            break
    assert feasible
    
    p_dl = sfc.power_alloc_qos_downlink(w, R, noise_pow, sinr_targets)
    margin_dl = sfc._sinr_margin_downlink(p_dl, w, R, noise_pow, sinr_targets)
    assert np.allclose(margin_dl, 0)

@hyp.settings(deadline=None)
@given(
    st.integers(min_value=2, max_value=5),
    st.integers(min_value=2, max_value=5),
)
def test_power_alloc_qos_zero_sinr_margin_uplink(num_zones, num_ls):
    """
    Can be flaky, because it does not always get a feasible problem to
        begin with. 
    """
    w = get_random_beamformer_normalized(num_zones, num_ls)
    R = get_random_spatial_cov(num_zones, num_ls)
    noise_pow = get_random_noise_pow(num_zones)
    rng = np.random.default_rng(SEED+12)
    sinr_targets = rng.uniform(low=0.05, high=0.05, size=num_zones)

    #R, noise_pow = sinropt.normalize_system(R, noise_pow)

    p = sfc._power_alloc_qos_uplink(w, R, noise_pow, sinr_targets)
    margin = sfc._sinr_margin_uplink(p, w,  R, noise_pow, sinr_targets)
    #sinr = sinropt._sinr_uplink_comparison(w, R, p, noise_pow)
    #sinr2 = sinropt.sinr_uplink(p, w, R, noise_pow)
    assert np.allclose(margin, 0)

    # p_dl = sinropt.power_alloc_qos_downlink(w, R, noise_pow, sinr_targets)
    # margin_dl = sinropt.sinr_margin_downlink(sinropt.apply_power_vec(w, p_dl), R, noise_pow, sinr_targets)
    
    # assert np.allclose(margin_dl, 0)

    


@hyp.settings(deadline=None)
@given(
    st.integers(min_value=2, max_value=5),
    st.integers(min_value=3, max_value=5),
    st.floats(min_value=1, max_value=10000),
)
def test_equal_capacity_uplink_downlink_duality(num_zones, num_ls, max_pow):
    w = get_random_beamformer_normalized(num_zones, num_ls)
    R = get_random_spatial_cov(num_zones, num_ls)
    ones = np.ones_like(get_random_noise_pow(num_zones))
    #noise_pow = get_random_noise_pow(num_zones)
    sinr_targets = get_random_sinr_targets(num_zones)

    q, c_ul = sfc._power_alloc_minmax_uplink(w, R, ones, sinr_targets, max_pow)
    p, c_dl = sfc._power_alloc_minmax_downlink(w, R, ones, sinr_targets, max_pow)
    assert np.allclose(c_ul, c_dl)


@hyp.settings(deadline=None)
@given(
    st.integers(min_value=2, max_value=5),
    st.integers(min_value=3, max_value=5),
)
def test_equal_feasibility_spectral_radius_uplink_downlink(num_zones, num_ls):

    w = get_random_beamformer_normalized(num_zones, num_ls)
    R = get_random_spatial_cov(num_zones, num_ls)
    sinr_targets = get_random_sinr_targets(num_zones)

    gain_ul = sfc._link_gain_uplink(w, R)
    spec_rad_ul = sfc._power_alloc_qos_feasibility_spectral_radius(gain_ul, sinr_targets)
    
    gain_dl = sfc._link_gain_downlink(w, R)
    spec_rad_dl = sfc._power_alloc_qos_feasibility_spectral_radius(gain_dl, sinr_targets)

    assert np.allclose(spec_rad_ul, spec_rad_dl)







@hyp.settings(deadline=None)
@given(
    st.integers(min_value=2, max_value=5),
    st.integers(min_value=3, max_value=5),
    #st.floats(min_value=100, max_value=10000),
)
def test_uplink_capacity_equals_feasiblity_spectral_radius(num_zones, num_ls):
    num_feasible = 0
    num_non_feasible = 0
    for i in range(1000):
        w = get_random_beamformer_normalized(num_zones, num_ls)
        R = get_random_spatial_cov(num_zones, num_ls)
        noise_pow = np.ones_like(get_random_noise_pow(num_zones))
        #noise_pow = get_random_noise_pow(num_zones)
        sinr_targets = get_random_sinr_targets(num_zones)

        gain_ul = sfc._link_gain_uplink(w, R)
        spec_rad_ul = sfc._power_alloc_qos_feasibility_spectral_radius(gain_ul, sinr_targets)
        
        q, c_ul = sfc._power_alloc_minmax_uplink(w, R, noise_pow, sinr_targets, 1e6)
        
        #print(c_ul > 1)
        if c_ul >= 1:
            assert spec_rad_ul < 1
            num_feasible += 1
        else:
            num_non_feasible += 1
    assert num_feasible > 0 
    assert num_non_feasible > 0
        #assert (c_ul >= 1) == (spec_rad_ul < 1)




@hyp.settings(deadline=None)
@given(
    st.integers(min_value=2, max_value=5),
    st.integers(min_value=5, max_value=10),
)
def test_strong_duality_holds_time_domain_sdr_vs_schubert(num_zones, bf_len):
    R = get_random_spatial_cov_td(num_zones, bf_len)
    noise_pow = np.ones(num_zones)
    sinr_targets = get_random_sinr_targets(num_zones)
    w, q = sfc.solve_power_weighted_qos_uplink(R, noise_pow, sinr_targets, 1e3, np.eye(bf_len)[None,:,:] * np.ones((num_zones, 1, 1))) #might be wrong dimension for audio cov

    audio_cov = np.stack([np.eye(bf_len) for _ in range(num_zones)], axis=0)
    
    W2 = sinr_constrained_weighted_pow_min_downlink(R, noise_pow, sinr_targets, audio_cov)
    w2 = sfc._select_solution_eigenvalue(W2)
    w2, p = sfc.extract_power_vec(w2)
    
    assert np.allclose(np.sum(q), np.sum(p))


@hyp.settings(deadline=None)
@given(
    st.integers(min_value=2, max_value=5),
    st.integers(min_value=3, max_value=10),
)
def test_strong_duality_holds_weighted_time_domain_sdr_vs_schubert(num_zones, bf_len):
    R = get_random_spatial_cov_td(num_zones, bf_len)
    audio_cov = get_random_audio_cov(num_zones, bf_len)
    noise_pow = np.ones(num_zones)
    sinr_targets = get_random_sinr_targets(num_zones)
    w, q = sfc.solve_power_weighted_qos_uplink(R, noise_pow, sinr_targets, 1e3, audio_cov)

    ul_objective = np.sum(q)
    
    W2 = sinr_constrained_weighted_pow_min_downlink(R, noise_pow, sinr_targets, audio_cov)
    w2 = sfc._select_solution_eigenvalue(W2)
    w2, p = sfc.extract_power_vec(w2)

    dl_objective = np.sum([p[k] * w2[k,:,None].T @ audio_cov[k,:,:] @ w2[k,:,None] for k in range(num_zones)])
    
    assert np.allclose(ul_objective, dl_objective)



@hyp.settings(deadline=None)
@given(
    st.integers(min_value=2, max_value=5),
    st.integers(min_value=2, max_value=5),
    st.floats(min_value=0.5, max_value=5),
)
def test_mwf_and_gevd_with_full_rank_are_equivalent(num_zones, num_ls, mu):
    Rb = get_random_spatial_cov_full_rank(num_zones, num_ls)
    Rd = get_random_spatial_cov_full_rank(num_zones, num_ls)

    mwf = sfc.szc_transform_mwf(Rb, Rd, mu)
    gevd = sfc.szc_transform_mwf_gevd(Rb, Rd, num_ls, mu)
    assert np.allclose(mwf, gevd)




"""Calculates distance measures for some different types of quantities. 

For all types of arrays, the mean square error can be computed. Vectors can be compared using angular distance or cosine similarity. For positive definite matrices, the correlation matrix distance, the affine invariant Riemannian metric, and the Kullback Leibler divergence between zero-mean Gaussian densities described by the compared matrices can be computed.

References
----------
[herdinCorrelation2005] M. Herdin, N. Czink, H. Ozcelik, and E. Bonek, 'Correlation matrix distance, a meaningful measure for evaluation of non-stationary MIMO channels,' in 2005 IEEE 61st Vehicular Technology Conference, May 2005, pp. 136-140 Vol. 1. doi: 10.1109/VETECS.2005.1543265. `[link] <https://doi.org/10.1109/VETECS.2005.1543265>`__ \n
[forstnermetric2003] W. Förstner and B. Moonen, 'A metric for covariance matrices,' in Geodesy-The Challenge of the 3rd Millennium, E. W. Grafarend, F. W. Krumm, and V. S. Schwarze, Eds., Berlin, Heidelberg: Springer Berlin Heidelberg, 2003, pp. 299–309. doi: 10.1007/978-3-662-05296-9_31. `[link] <https://doi.org/10.1007/978-3-662-05296-9_31>`__ \n
[duchiDerivations2016] J. Duchi, 'Derivations for Linear Algebra and Optimization, 2016. `[link] <https://web.stanford.edu/~jduchi/projects/general_notes.pdf>`__ \n
[absilOptimization2008] P.-A. Absil, R. Mahony, and R. Sepulchre, Optimization algorithms on matrix manifolds. Princeton, N.J. ; Woodstock: Princeton University Press, 2008. \n
"""
import numpy as np
import scipy.linalg as splin


def mse(var1, var2):
    """The normalized mean square error

    Normalized by the second variable

    Parameters
    ----------
    var1 : np.ndarray of any shape
        First variable
    var2 : np.ndarray of the same shape as var1
        Second variable. Cannot be zero, as it is used as the denominator
    
    Returns
    -------
    mse : float
        The normalized mean square error    
    """
    return np.sum(np.abs(var1 - var2)**2) / np.sum(np.abs(var2)**2)



#============== FOR VECTORS ======================
def angular_distance(vec1, vec2, sign_invariant=False):
    """A distance metric based on the cosine similary, that retains the
        scale invariant property, but is also a proper distance metric

    Parameters
    ----------
    vec1 : np.ndarray of shape (N,)
        First vector
    vec2 : np.ndarray of shape (N,)
        Second vector
    sign_invariant : bool, optional
        if True, the angle is first adjusted to a range between 1 and 0
        meaning that parallell vectors and opposite vectors are both considered to be the same
        If False, the same shape but opposite signs gives maximum distance

    Returns
    -------
    ang_dist : float
        The angular distance between the two vectors, in the range [0, 1]
    """
    sim = cos_similary(vec1, vec2)
    if sign_invariant:
        sim = np.abs(sim)
    return np.arccos(sim) / np.pi

def cos_similary(vec1, vec2):
    """Computes <vec1, vec2> / (||vec1|| ||vec2||) which is cosine of the angle between the two vectors. 
        1 is paralell vectors, 0 is orthogonal, and -1 is opposite directions

    Parameters
    ----------
    vec1 : np.ndarray of any shape
        First vector. If the arrays have more than 1 axis, it will be flattened.
    vec2 : np.ndarray of same shape as vec1
        Second vector. If the arrays have more than 1 axis, it will be flattened.

    Returns
    -------
    cos_sim : float
        The cosine similarity between the two vectors
    """
    assert vec1.shape == vec2.shape
    vec1 = np.ravel(vec1)
    vec2 = np.ravel(vec2)
    norms = np.linalg.norm(vec1) *np.linalg.norm(vec2)
    if norms == 0:
        return np.nan
    ip = vec1.T @ vec2
    return ip / norms

def spatial_similarity(vec1, vec2):
    """Measures the spatial similarity between two vectors. Also known as the modal assurance criterion (MAC).

    1 is identical, 0 is fully dissimilar

    Implements |p^H q|^2 / (||p||^2 ||q||^2)
    
    Parameters
    ----------
    vec1 : np.ndarray of shape (..., N)
        First vector
    vec2 : np.ndarray of shape (..., N)
        Second vector

    Returns
    -------
    sim : float or ndarray of shape (...)
        The spatial similarity between the two vectors
    
    References
    ----------
    (25) in M. Hahmann and E. Fernandez-Grande, “A convolutional plane wave model for sound field reconstruction.” Aug. 24, 2022.
    """
    assert vec1.shape == vec2.shape

    denom = np.linalg.norm(vec1, axis=-1)**2 * np.linalg.norm(vec2, axis=-1)**2
    return np.abs(np.sum(vec1.conj() * vec2, axis=-1))**2 / denom
    #norm2 = np.linalg.norm(vec2, ord=2, axis=-1)**2



#=============== FOR COVARIANCE MATRICES =============

def corr_matrix_distance(mat1, mat2):
    """Computes the correlation matrix distance
    
    0 means that the matrices are equal up to a scaling
    1 means that they are maximally different (orthogonal in NxN dimensional space)

    Parameters
    ----------
    mat1 : np.ndarray of shape (..., N, N)
        First covariance matrix, should be symmetric and positive definite
    mat2 : np.ndarray of shape (..., N, N)
        Second covariance matrix, should be symmetric and positive definite

    References
    ----------
    Correlation matrix distaince, a meaningful measure for evaluation of 
    non-stationary MIMO channels - Herdin, Czink, Ozcelik, Bonek
    """
    assert mat1.shape == mat2.shape
    norm1 = np.linalg.norm(mat1, ord="fro", axis=(-2,-1))
    norm2 = np.linalg.norm(mat2, ord="fro", axis=(-2,-1))
    if norm1 * norm2 == 0:
        return np.array(np.nan)
    return np.real_if_close(1 - np.trace(mat1 @ mat2) / (norm1 * norm2))


def covariance_distance_riemannian(mat1, mat2):
    """
    Computes the covariance matrix distance

    Parameters
    ----------
    mat1 : np.ndarray of shape (N, N)
        First covariance matrix, should be symmetric and positive definite
    mat2 : np.ndarray of shape (N, N)
        Second covariance matrix, should be symmetric and positive definite

    Returns
    -------
    dist : float
        The distance between the two matrices
    
    Notes
    -----
    It is the distance of a canonical invariant Riemannian metric on the space 
    Sym+(n, R) of real symmetric positive definite matrices. 

    Invariant to affine transformations and inversions. 
    It is a distance measure, so 0 means equal and then it goes to infinity
    and the matrices become more unequal.

    When the metric of the space is the fisher information metric, this is the 
    distance of the space. See COVARIANCE CLUSTERING ON RIEMANNIAN MANIFOLDS
    FOR ACOUSTIC MODEL COMPRESSION - Shinohara, Masukp, Akamine

    References
    ----------
    [forstnermetric2003]
    [absilOptimization2008]
    """
    assert mat1.shape == mat2.shape
    assert mat1.shape[0] == mat1.shape[1]
    assert mat1.ndim == 2
    eigvals = splin.eigh(mat1, mat2, eigvals_only=True)
    return np.real_if_close(np.sqrt(np.sum(np.log(eigvals)**2)))


def covariance_distance_kl_divergence(mat1, mat2):
    """The Kullback Leibler divergence between two Gaussian
    distributions that has mat1 and mat2 as their covariance matrices. 

    Assumes both of these distributions has zero mean.

    It is a distance measure, so 0 means equal and then it goes to infinity
    and the matrices become more unequal.

    Parameters
    ----------
    mat1 : np.ndarray of shape (N, N)
        First covariance matrix, should be symmetric and positive definite
    mat2 : np.ndarray of shape (N, N)
        Second covariance matrix, should be symmetric and positive definite
    
    Returns
    -------
    dist : float
        The distance between the two matrices
    
    """
    assert mat1.shape == mat2.shape
    assert mat1.shape[0] == mat1.shape[1]
    assert mat1.ndim == 2
    N = mat1.shape[0]
    eigvals = splin.eigh(mat1, mat2, eigvals_only=True)

    det1 = splin.det(mat1)
    det2 = splin.det(mat2)
    common_trace = np.sum(eigvals)
    return np.real_if_close(np.sqrt((np.log(det2 / det1) + common_trace - N) / 2))




"""Module for plotting helper functions

References
----------
"""

import matplotlib.pyplot as plt

try:
    import tikzplotlib
except ImportError:
    tikzplotlib = None

def _tikzplotlib_fix_ncols(obj):
    """Workaround for matplotlib 3.6 renamed legend's _ncol to _ncols, which breaks tikzplotlib

    Parameters
    ----------
    obj : Figure object
        Figure object to fix the _ncol attribute in.
    """
    if hasattr(obj, "_ncols"):
        obj._ncol = obj._ncols
    for child in obj.get_children():
        _tikzplotlib_fix_ncols(child)


def save_plot(print_method, folder, name=""):
    """Save plot to file in a number of formats.

    Parameters
    ----------
    print_method : str
        Method for saving the plot. Options are 'show', 'tikz', 'pdf', 'svg', 'none'.
        If 'show', the plot is shown in a window.
        If 'tikz', the plot is saved as a tikz file and a pdf file. Requires tikzplotlib installed. 
        If 'pdf', the plot is saved as a pdf file.
        If 'svg', the plot is saved as a svg file.
        If 'none', the plot is not saved.
    folder : Path
        Folder to save the plot in.
    name : str, optional
        Name of the file. The default is "".
    """
    if print_method == "show":
        plt.show()
    elif print_method == "tikz":
        if folder is not None:
            nested_folder = folder.joinpath(name)
            try:
                nested_folder.mkdir()
            except FileExistsError:
                pass

            fig = plt.gcf()
            _tikzplotlib_fix_ncols(fig)
            tikzplotlib.save(
                str(nested_folder.joinpath(f"{name}.tex")),
                externalize_tables=True,
                float_format=".8g",
            )
            plt.savefig(
                str(nested_folder.joinpath(name + ".pdf")),
                dpi=300,
                facecolor="w",
                edgecolor="w",
                orientation="portrait",
                format="pdf",
                transparent=True,
                bbox_inches=None,
                pad_inches=0.2,
            )
    elif print_method == "pdf":
        if folder is not None:
            plt.savefig(
                str(folder.joinpath(name + ".pdf")),
                dpi=300,
                facecolor="w",
                edgecolor="w",
                orientation="portrait",
                format="pdf",
                transparent=True,
                bbox_inches="tight",
                pad_inches=0.2,
            )
    elif print_method == "svg":
        if folder is not None:
            plt.savefig(
                str(folder.joinpath(name + ".svg")),
                dpi=300,
                format="svg",
                transparent=True,
                bbox_inches="tight",
                pad_inches=0.2,
            )
    elif print_method == "none":
        pass
    else:
        raise ValueError
    plt.close("all")


def set_basic_plot_look(ax):
    """Sets basic look for a plot.
    
    Parameters
    ----------
    ax : Axes
        Axes object to set the look of.
    """
    ax.grid(True)
    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)


"""Functions for handling the spherical harmonic wave function and estimating the sound field coefficients

All math and translation theorems are taken from [martinMultiple2006] unless otherwise stated.

References
----------
[martinMultiple2006] P. A. Martin, Multiple scattering: Interaction of time-harmonic waves with N obstacles, vol. 107. in Encyclopedia of mathematics and its applications, vol. 107. Cambridge, UK: Cambridge University Press, 2006. \n
[brunnstromBayesianSubmitted] J. Brunnström, M. B. Møller, and M. Moonen, “Bayesian sound field estimation using moving microphones,” IEEE Open Journal of Signal Processing, submitted. \n
[uenoSound2018] N. Ueno, S. Koyama, and H. Saruwatari, “Sound field recording using distributed microphones based on harmonic analysis of infinite order,” IEEE Signal Process. Lett., vol. 25, no. 1, pp. 135–139, Jan. 2018, doi: 10.1109/LSP.2017.2775242. \n


Equivalence of spherical harmonic definitions
---------------------------------------------
The scipy spherical harmonic implementation is 
https://docs.scipy.org/doc/scipy/reference/generated/scipy.special.sph_harm.html

The Legendre polynomial (referred to as Legendre function of the first kind in Scipy documentation)
definition given in Scipy is (37) after inserting the definition of the Hypergeometric function. 
The definition used in Martin is (30), the Rodrigues formula.
The two definitions are equivalent, therefore the Legendre polynomial is consistent. 
https://mathworld.wolfram.com/LegendrePolynomial.html

Martins definition of the Associated Legendre Polynomial $P_n^m(t)$, found in (A.1) in [martinMultiple2006]
is $P_n^m(t) = (1-t^2)^{m/2} \\frac{d^m}{dt^m} P_n(t)$. Scipy has almost the same definition, $S_n^m(t) = (-1)^m P_n^m(t)$,
where S_n^m(t) denotes the Scipy definition. Therefore they are equivalent except for that Scipy includes
the Condon-Shortley phase in the ALP. 

The definition of the spherical harmonics in (3.6) in [martinMultiple2006] is equivalent to Scipy's definition,
except for the inclusion of an additional factor (-1)^m, i.e. the Condon-Shortley phase. Because the 
same factor is included in Scipy's ALP but not Martin ALP, the spherical harmonic definitions are equivalent. 
"""
import numpy as np
import itertools as it
import scipy.special as special

import aspcol.kernelinterpolation as ki
import aspcol.utilities as utils

import wigners


# ==================== BASIC SHD FUNCTIONALITY ====================

def shd_max_order(num_coeffs):
    """Returns the maximum order of the spherical harmonics that is required to represent a given number of coefficients

    Parameters
    ----------
    num_coeffs : int
        number of coefficients

    Returns
    -------
    max_order : int
        maximum order of the spherical harmonics that is required to represent the coefficients
    """
    order = np.sqrt(num_coeffs) - 1
    assert order == int(order)
    return int(order)

def shd_num_coeffs(max_order):
    """Number of coefficients required to represent a spherical harmonic function of a given order
    
    Parameters
    ----------
    max_order : int
        maximum order of the spherical harmonics
    
    Returns
    -------
    num_coeffs : int
        number of coefficients required to represent a spherical harmonic function of the given order
    """
    return int((max_order+1)**2)


def shd_num_degrees(max_order : int):
    """
    Returns a list of mode indices for each order
    when order = n, the degrees are only non-zero for -n <= degree <= n

    Parameters
    ----------
    max_order : int
        is the maximum order that is included

    Returns
    -------
    degree : list of ndarrays of shape (2*order+1)
        so the ndarrays will grow larger for higher list indices
    """
    degree = []
    for n in range(max_order+1):
        pos_degrees = np.arange(n+1)
        degree_n = np.concatenate((-np.flip(pos_degrees[1:]), pos_degrees))
        degree.append(degree_n)
    return degree

def shd_num_degrees_vector(max_order : int):
    """
    Constructs a vector with the index of each order and degree
    when order = n, the degrees are only non-zero for -n <= degree <= n

    Parameters
    ----------
    max_order : int
        is the maximum order that is included

    Returns
    -------
    order : ndarray of shape ()
    degree : ndarray of shape ()
    """
    order = []
    degree = []

    for n in range(max_order+1):
        pos_degrees = np.arange(n+1)
        degree_n = np.concatenate((-np.flip(pos_degrees[1:]), pos_degrees))
        degree.append(degree_n)

        order.append(n*np.ones_like(degree_n))
    degree = np.concatenate(degree)
    order = np.concatenate(order)
    return order, degree

def shd_min_order(wave_num, radius):
    """
    Returns the minimum order of the spherical harmonics that should be used
    for a given wavenumber and radius

    Here according to the definition in Katzberg et al., Spherical harmonic
    representation for dynamic sound-field measurements

    Parameters
    ----------
    wave_num : ndarray of shape (num_freqs)
    radius : float
        represents r_max in the Katzberg paper

    Returns
    -------
    M_f : ndarray of shape (num_freqs)
        contains an integer which is the minimum order of the spherical harmonics
    """
    return np.ceil(wave_num * radius).astype(int)

def shd_coeffs_order_degree_to_index(order, degree):
    return order**2 + order + degree





def shd_basis(pos, order, degree, wavenum):
    """Spherical harmonic basis function for sound field in 3D

    Implements: sqrt(4pi) j_order(kr) Y_order^degree(polar angle, zenith angle)
    degree and order of the spherical harmonics might be swapped according to some
    definitions.
    
    Parameters
    ----------
    pos : ndarray of shape (num_pos, 3)
        positions of the evaluation points
    order : ndarray of shape (num_coeffs,)
        order of the spherical harmonics. Can be any non-negative integer
    degree : ndarray of shape (num_coeffs,)
        degree of the spherical harmonics. Must satisfy |degree[i]| <= order[i] for all i
    wave_num : float or ndarray of shape (num_freq,)
        wavenumber, defined as w / c where w is the angular 
        frequency and c is the speed of sound

    Returns
    -------
    f : ndarray of shape (num_freq, num_coeffs, num_pos)
        values of the spherical harmonic basis function at the evaluation points

    Notes
    -----
    See more detailed derivations in any of:
    [brunnstromBayesianSubmitted] J. Brunnström, M. B. Møller, and M. Moonen, “Bayesian sound field estimation using moving microphones,” IEEE Open Journal of Signal Processing, submitted.
    [uenoSound2018] N. Ueno, S. Koyama, and H. Saruwatari, “Sound field recording using distributed microphones based on harmonic analysis of infinite order,” IEEE Signal Process. Lett., vol. 25, no. 1, pp. 135–139, Jan. 2018, doi: 10.1109/LSP.2017.2775242.
    """
    #Parse arguments to get correct shapes
    radius, angles = utils.cart2spherical(pos)
    radius = radius[None, None, ...]
    angles = angles[None, None, ...]

    order = order[None, :, None]
    degree = degree[None, :, None]

    if isinstance(wavenum, (int, float)):
        wavenum = np.array([wavenum])
    assert wavenum.ndim == 1
    wavenum = wavenum[:, None, None]

    # Calculate the function values
    f = np.sqrt(4*np.pi) * special.spherical_jn(order, wavenum * radius) * special.sph_harm(degree, order, angles[...,0], angles[...,1])
    return f


def reconstruct_pressure(shd_coeffs, pos, expansion_center, wavenum):
    """Returns the complex sound pressure at some positions using the provided spherical harmonic coefficients
    
    Parameters
    ----------
    shd_coeffs : ndarray of shape (num_freq, num_coeffs,)
        spherical harmonic coefficients of the sound field
    pos : ndarray of shape (num_pos, 3)
        positions of the evaluation points
    expansion_center : ndarray of shape (1, 3)
        expansion center of the spherical harmonics
    max_order : int
        maximum order of the spherical harmonics. Could be removed in a future version of 
        the function, but must currently equal the max_order of the provided coefficients
    wavenum : ndarray of shape (num_freq,)
        wavenumber, defined as w / c where w is the angular frequency
        and c is the speed of sound. 

    Returns
    -------
    p_est : ndarray of shape (num_freq, num_pos,)
        complex sound pressure at the evaluation points
    """
    if not shd_coeffs.ndim == 2:
        shd_coeffs = shd_coeffs[None, ...]
        wavenum = np.array([wavenum])
    assert shd_coeffs.ndim == 2
    assert wavenum.ndim == 1
    assert shd_coeffs.shape[0] == wavenum.shape[0]

    num_coeffs = shd_coeffs.shape[1]
    max_order = shd_max_order(num_coeffs)

    rel_pos = pos - expansion_center
    order, degree = shd_num_degrees_vector(max_order)
    basis_values = shd_basis(rel_pos, order, degree, wavenum)
    p_est = np.sum(shd_coeffs[...,None] * basis_values, axis=1)
    return p_est













# ==================== TRANSLATION OF SHD COEFFICIENTS ====================
def translate_shd_coeffs(shd_coeffs, pos, wave_num, max_order_output):
    """Translates the provided shd_coeffs to another expansion center
    
    shd_coeffs(pos_orig + pos) = translate(pos) shd_coeffs(pos_orig)
    
    Parameters
    ----------
    shd_coeffs : ndarray of shape (num_freqs, num_coeffs,) or (num_freqs, num_pos, num_coeffs,)
        coefficients of the sequence
    pos : ndarray of shape (num_pos, 3)
        position argument to the translation operator
    wave_num : ndarray of shape (num_freq,)
        wavenumber, defined as w / c where w is the angular frequency
        and c is the speed of sound.
    max_order_output : int
        maximum order of the translated coefficients

    Returns
    -------
    shd_coeffs_translated : ndarray of shape (num_freqs, num_pos, num_coeffs,)
        translated coefficients
        if num_pos == 1, the returned array will have shape (num_freqs, num_coeffs,)
    """
    if shd_coeffs.ndim == 2:
        shd_coeffs = shd_coeffs[:,None,:]
    assert pos.ndim == 2
    assert pos.shape[1] == 3
    assert wave_num.ndim == 1
    assert shd_coeffs.ndim == 3
    assert shd_coeffs.shape[0] == wave_num.shape[0] or shd_coeffs.shape[0] == 1
    assert shd_coeffs.shape[1] == pos.shape[0] or shd_coeffs.shape[1] == 1
    num_coeffs = shd_coeffs.shape[-1]

    max_order_input = shd_max_order(num_coeffs)

    T_all = translation_operator(pos, wave_num, max_order_input, max_order_output)
    translated_coeffs = T_all @ shd_coeffs[...,None]

    num_pos = pos.shape[0]
    if num_pos == 1:
        translated_coeffs = np.squeeze(translated_coeffs, axis=1)
    return np.squeeze(translated_coeffs, axis=-1)

def translation_operator(pos, wave_num, max_order_input, max_order_output):
    """Translation operator for harmonic coefficients, such that 
    shd_coeffs(pos_orig + pos) = T(pos) @ shd_coeffs(pos_orig)

    Defined according to $T = \\hat{S}^T $, where $\\hat{S}$ is the basis translation matrix defined in [martinMultiple2006].
    This definition is consistent with the definition (6) in [uenoSound2018]. 
    
    Parameters
    ----------
    pos : ndarray of shape (num_pos, 3)
        position argument to the translation operator
    wave_num : ndarray of shape (num_freq,)
        wavenumber, defined as w / c where w is the angular frequency
        and c is the speed of sound.
    max_order_input : int
        maximum order of the coefficients that should be translated
    max_order_output : int
        maximum order of the translated coefficients

    Returns
    -------
    T : ndarray of shape (num_freqs, num_pos, num_coeffs_output, num_coeffs_input)
        translation operator, such that shd_coeffs(pos_orig + pos) = T(pos) @ shd_coeffs(pos_orig)

    References
    ----------
    [martinMultiple2006] P. A. Martin, Multiple scattering: Interaction of time-harmonic waves with N obstacles, vol. 107. in Encyclopedia of mathematics and its applications, vol. 107. Cambridge, UK: Cambridge University Press, 2006.
    [uenoSound2018] N. Ueno, S. Koyama, and H. Saruwatari, “Sound field recording using distributed microphones based on harmonic analysis of infinite order,” IEEE Signal Process. Lett., vol. 25, no. 1, pp. 135–139, Jan. 2018, doi: 10.1109/LSP.2017.2775242.
    """

    S = basis_translation_3_80(pos, wave_num, max_order_output, max_order_input)
    T = np.moveaxis(S, -1, -2)
    return T

def basis_translation_3_80(pos, wave_num, max_order_input, max_order_output):
    """Translation operator for shd basis function, such that 
    shd_basis(pos_orig + pos) = T(pos) @ shd_basis(pos_orig)x

    Implemented according to equation 3.80 in 
    P. Martin, Multiple scattering: Interaction of time-harmonic waves with N obstacles.
    
    Parameters
    ----------
    pos : ndarray of shape (num_pos, 3)
        position argument to the translation operator
    wave_num : ndarray of shape (num_freq,)
        wavenumber, defined as w / c where w is the angular frequency
        and c is the speed of sound.
    max_order_input : int
        maximum order of the coefficients that should be translated
    max_order_output : int
        maximum order of the translated coefficients

    Returns
    -------
    T : ndarray of shape (num_freqs, num_pos, num_coeffs_output, num_coeffs_input)
        translation operator, such that shd_basis(pos_orig + pos) = T(pos) @ shd_basis(pos_orig)
    """
    num_coeffs_input = shd_num_coeffs(max_order_input)
    num_coeffs_output = shd_num_coeffs(max_order_output)
    num_pos = pos.shape[0]
    num_freq = wave_num.shape[0]

    tr_op = np.zeros((num_freq, num_pos, num_coeffs_output, num_coeffs_input), dtype=complex)
    orders_input, degrees_input = shd_num_degrees_vector(max_order_input)
    orders_output, degrees_output = shd_num_degrees_vector(max_order_output)

    gaunt_set = _calculate_gaunt_set(max_order_input, max_order_output)

    max_tot_order = int(max_order_input + max_order_output)
    orders_tot, degrees_tot = shd_num_degrees_vector(max_tot_order)
    all_q = np.arange(max_tot_order+1)
    #nu_plus_n = orders_output[:,None] + orders_input[None,:]
    #mu_minus_m = degrees_output[:,None] - degrees_input[None,:]
    # all_q = np.arange(nu_plus_n)[nu_plus_n >= all_q]
    
    radius, angles = utils.cart2spherical(pos)
    radius = radius[None, None, ...]
    angles = angles[None, ...]

    orders_tot = orders_tot[:, None]
    degrees_tot = degrees_tot[:, None]

    wave_num = wave_num[:, None, None]
    all_q = all_q[None,:,None]

    bessel = special.spherical_jn(all_q, wave_num * radius) # ordered as (num_freq, num_q, num_pos)
    harm = np.conj(special.sph_harm(degrees_tot, orders_tot, angles[...,0], angles[...,1])) # ordered as (num_coeffs_tot, num_pos)

    for out_idx, (n, m) in enumerate(zip(orders_output, degrees_output)):
        for in_idx, (nu, mu) in enumerate(zip(orders_input, degrees_input)):
            q_array = np.arange(np.abs(mu-m), n+nu+1)
            harm_idxs = shd_coeffs_order_degree_to_index(q_array, mu-m)

            basis_val = bessel[:,q_array,:] * harm[None,harm_idxs, :]
            g = (1j)**q_array[None,:,None] * gaunt_set[out_idx, in_idx, q_array][None,:,None]
            factor = 4*np.pi * (1j)**(nu-n) * (-1.0)**(m)
            sum_val = factor * np.sum(basis_val * g, axis=1)
            tr_op[..., out_idx, in_idx] = sum_val

    for p in range(num_pos):
        if np.sum(np.abs(pos[p,:])) == 0: 
            tr_op[:, p, :, :] = np.eye(num_coeffs_output, num_coeffs_input)[None,:,:]
    return tr_op

def _calculate_gaunt_set(max_order_input, max_order_output):
    """Numpy function for precaculating the Gaunt coefficients for the translation operator.

    Used to avoid having to write a jax implementation of wigner coefficients. 
    """
    num_coeffs_input = shd_num_coeffs(max_order_input)
    num_coeffs_output = shd_num_coeffs(max_order_output)
    orders_input, degrees_input = shd_num_degrees_vector(max_order_input)
    orders_output, degrees_output = shd_num_degrees_vector(max_order_output)
    gaunt = np.zeros((num_coeffs_output, num_coeffs_input, max_order_input+max_order_output+1), dtype=float)
    for out_idx, (n, m) in enumerate(zip(orders_output, degrees_output)):
        for in_idx, (nu, mu) in enumerate(zip(orders_input, degrees_input)):
            for q in range(n+nu+1):
                if np.abs(mu-m) <= q:
                    g = gaunt_coefficient(n, m, nu, -mu, q)
                    gaunt[out_idx, in_idx, q] = g
    return gaunt


def basis_translation_3_92(pos, wave_num, max_order_input, max_order_output):
    """Translation operator for harmonic basis functions, such that 
    shd_basis(pos_orig + pos) = T(pos) @ shd_basis(pos_orig)

    Taken directly from 3.92 in P. Martin, Multiple scattering: Interaction of 
    time-harmonic waves with N obstacles

    
    Parameters
    ----------
    pos : ndarray of shape (num_pos, 3)
        position argument to the translation operator
    wave_num : ndarray of shape (num_freq,)
        wavenumber, defined as w / c where w is the angular frequency
        and c is the speed of sound.
    max_order_input : int
        maximum order of the coefficients that should be translated
    max_order_output : int
        maximum order of the translated coefficients

    Returns
    -------
    T : ndarray of shape (num_freqs, num_pos, num_coeffs_output, num_coeffs_input)
        translation operator, such that shd_coeffs(pos_orig + pos) = T(pos) @ shd_coeffs(pos_orig)
    """
    num_coeffs_input = shd_num_coeffs(max_order_input)
    num_coeffs_output = shd_num_coeffs(max_order_output)
    num_pos = pos.shape[0]
    num_freq = wave_num.shape[0]

    tr_op = np.zeros((num_freq, num_pos, num_coeffs_output, num_coeffs_input), dtype=complex)

    orders_input, degrees_input = shd_num_degrees_vector(max_order_input)
    orders_output, degrees_output = shd_num_degrees_vector(max_order_output)
    for f in range(num_freq):
        for p in range(num_pos):
            if np.sum(np.abs(pos[p,:])) == 0:
                tr_op[f, p, :, :] = np.eye(num_coeffs_output, num_coeffs_input)
            else:
                for out_idx, (n, m) in enumerate(zip(orders_output, degrees_output)):
                    for in_idx, (nu, mu) in enumerate(zip(orders_input, degrees_input)):
                        sum_val = 0
                        q0 = _linearity_formula_lower_limit(n, nu, m, -mu)
                        Q = (n + nu - q0) / 2
                        assert Q == int(Q)
                        Q = int(Q)
                        for q in range(Q+1):
                        #for q in range(n+nu+1):
                            if np.abs(m-mu) <= (q0+2*q):
                                basis_val = shd_basis(pos[p:p+1,:], np.array([q0+2*q]), np.array([m-mu]), wave_num[f])
                                #basis_val = shd_basis(pos[p:p+1,:], np.array([q]), np.array([mu-m]), wave_num[f])
                                g = gaunt_coefficient(n, m, nu, -mu, q0+2*q)
                                sum_val += (-1)**q * basis_val * g
                        sum_val *= np.sqrt(4*np.pi) * (-1)**(nu + mu + Q)
                        tr_op[f, p, out_idx, in_idx] = np.squeeze(sum_val)
    return tr_op

def _linearity_formula_lower_limit(l1, l2, m1, m2):
    """ Lower limit of the sum in the linearity formula for the Gaunt coefficient
    described in (3.74) in Multiple scattering: Interaction of time-harmonic waves with N obstacles.

    func(l1, l2, m1, m2) is equivalent to func(l1, l1, m2, m1)
    """
    if np.abs(l1 - l2) >= np.abs(m1 + m2):
        return np.abs(l1 - l2)
    elif np.abs(l1 - l2) < np.abs(m1 + m2) and ((l1 + l2 + np.abs(m1 + m2)) % 2 == 0):
        return np.abs(m1 + m2)
    elif np.abs(l1 - l2) < np.abs(m1 + m2) and ((l1 + l2 + np.abs(m1 + m2)) % 2 == 1):
        return np.abs(m1 + m2) + 1
    else:
        raise ValueError("Something went wrong")




def gaunt_coefficient(l1, m1, l2, m2, l3):
    """Gaunt coefficient G(l1, m1, l2, m2, l3)
    
    Defined on page 83, equation (3.71) in [martinMultiple2006]. Argument order is the same as in the reference
    
    Parameters
    ----------
    l1 : int
        Spherical harmonic order
    m1 : int
        Spherical harmonic degree
    l2 : int
        Spherical harmonic order
    m2 : int
        Spherical harmonic degree
    l3 : int
        Spherical harmonic order

    Returns
    -------
    G : float
        Gaunt coefficient for the given arguments

    References
    ----------
    [martinMultiple2006] P. A. Martin, Multiple scattering: Interaction of time-harmonic waves with N obstacles, vol. 107. in Encyclopedia of mathematics and its applications, vol. 107. Cambridge, UK: Cambridge University Press, 2006.
    
    Notes
    -----
    The relationship between this and triple_harmonic_integral 
    (the latter is the definition of gaunt coefficient given by Sympy)
    is I(l1, l2, l3, m1, m2, m3) = delta(m1+m2+m3,0) * (-1)**(m3) * gaunt(l1, m1, l2, m2, l3)
    
    Can be seen on the final equation on page 328 in Multiple scattering: Interaction of time-harmonic waves with N obstacles
    by P. A. Martin, 2006. 

    A recursive algorithm can be found in:
    Fast evaluation of Gaunt coefficients: recursive approach - Yu-lin Xu, 1997
    """
    f1 = (-1.0)**(m1+m2)
    f2 = np.sqrt((2*l1+1)*(2*l2+1)*(2*l3+1)/(4*np.pi))
    f3 = wigners.wigner_3j(l1,l2,l3,m1,m2,-m1-m2)
    f4 = wigners.wigner_3j(l1,l2,l3,0,0,0)
    return f1 * f2 * f3 * f4

def triple_harmonic_integral(l1, l2, l3, m1, m2, m3):
    """Integral of the product of three spherical harmonics

    Defined as $\\int_{\\omega} Y_{l1,m1}(\\omega) Y_{l2,m2}(\\omega) Y_{l3,m3}(\\omega) d\\omega$
    where $\\omega$ is the angle. It is sometimes (in Sympy for example) called gaunt coefficient.
    
    Parameters
    ----------
    l1, l2, l3 : int
        Spherical harmonic orders
    m1, m2, m3 : int
        Spherical harmonic degrees

    Returns
    -------
    integral_value : float
        the value of the triple harmonic integral
    
    """
    f1 = np.sqrt((2*l1+1)*(2*l2+1)*(2*l3+1)/(4*np.pi))
    f2 = wigners.wigner_3j(l1,l2,l3,m1,m2,m3)
    f3 = wigners.wigner_3j(l1,l2,l3,0,0,0)
    return f1 * f2 * f3

def gaunt_coef_lookup(max_order):
    """
    Returns a lookup table for the gaunt coefficients
    """
    all_params = np.array([i for i in _all_gaunt_parameters(max_order)])

    tbl = np.zeros((max_order, 2*max_order+1, max_order, 2*max_order+1, max_order))

    val_ref = []
    for i in range(all_params.shape[0]):
        (l1, m1, l2, m2, l3) = all_params[i]
        tbl[l1, m1, l2, m2, l3] = gaunt_coefficient(l1, m1, l2, m2, l3)
    val_ref = np.array(val_ref)
    return tbl

def _all_gaunt_parameters(max_order):
    """ This might not include all parameters"""
    all_orders = it.product(range(max_order), range(max_order), range(max_order))
    for l1, l2, l3 in all_orders:
        all_modes = it.product(range(0, l1), range(0, l2))
        for m1, m2 in all_modes:
            if abs(l1-l2) <= l3 <= l1+l2:
                if abs(m1 + m2) <= l3:
                    if (l1 + l2 + l3) % 2 == 0:
                        yield l1, m1, l2, m2, l3






# =================== DIRECTIVITY DEFINITIONS ===================

def directivity_omni(max_order = 0):
    """Harmonic coefficients of an omnidirectional directivity function

    Parameters
    ----------
    max_order : int
        maximum order of the spherical harmonics. If set higher than 0, the coefficients of the higher orders will be zero.

    Returns
    -------
    dir_coeffs : ndarray of shape (1, num_coeffs)
        coefficients of the directivity function

    References
    ----------
    [uenoSound2018] N. Ueno, S. Koyama, and H. Saruwatari, “Sound field recording using distributed microphones based on harmonic analysis of infinite order,” IEEE Signal Process. Lett., vol. 25, no. 1, pp. 135–139, Jan. 2018, doi: 10.1109/LSP.2017.2775242.
    [brunnstromBayesianSubmitted] J. Brunnström, M. B. Møller, and M. Moonen, “Bayesian sound field estimation using moving microphones,” IEEE Open Journal of Signal Processing, submitted.

    Notes
    -----
    The directivity coefficients are defined in more detail in the supplementary notes of [uenoSound2018] and in [brunnstromBayesianSubmitted]
    """
    if max_order == 0:
        return np.ones((1,1), complex)
    else:
        num_coeffs = shd_num_coeffs(max_order)
        dir = np.zeros((1, num_coeffs), dtype=complex)
        dir[0,0] = 1
        return dir
    
def directivity_linear(A, d_mic, max_order = 1):
    """Harmonic coefficients of a linear directivity function

    The directivity function is defined as
    c(d, omega) = A + A * d^T d_mic

    Omni directivity is obtained by setting A = 0
    Cardoid directivity is obtained by setting A = 1/2
    Figure-8 directivity is obtained by setting A = 1
    
    Parameters
    ----------
    A : float
        directivity parameter. Must be between 0 and 1. 
    d_mic : ndarray of shape (num_mic, 3)
        direction of the microphone(s). It is the angle at which a cardioid microphone would give the strongest response
        It is the outward direction, so from the center of the microphone towards the sources
        Must be a unit vector
    max_order : int
        maximum order of the spherical harmonics. If set higher than 1, 
        the coefficients of the higher orders will be zero. 
    
    Returns
    -------
    dir_coeffs : ndarray of shape (num_mic, num_coeffs)
        coefficients of the directivity function

    References
    ----------
    [uenoSound2018] N. Ueno, S. Koyama, and H. Saruwatari, “Sound field recording using distributed microphones based on harmonic analysis of infinite order,” IEEE Signal Process. Lett., vol. 25, no. 1, pp. 135–139, Jan. 2018, doi: 10.1109/LSP.2017.2775242.
    [brunnstromBayesianSubmitted] J. Brunnström, M. B. Møller, and M. Moonen, “Bayesian sound field estimation using moving microphones,” IEEE Open Journal of Signal Processing, submitted.

    Notes
    -----
    The directivity coefficients are defined in more detail in the supplementary notes of [uenoSound2018] and in [brunnstromBayesianSubmitted]
    """
    num_mic = d_mic.shape[0]
    assert d_mic.shape == (num_mic, 3)
    assert max_order >= 1
    assert 0 <= A <= 1
    radius, angles = utils.cart2spherical(d_mic)
    assert np.allclose(radius, 1)
    
    dir_coeffs = np.zeros((num_mic, shd_num_coeffs(max_order)), dtype=complex)
    order, degree = shd_num_degrees_vector(max_order) 

    dir_coeffs[:, order == 0] = 1 - A
    degrees_to_set = degree[order == 1]
    harm = special.sph_harm(degrees_to_set[None,:], 1, angles[...,0:1], angles[...,1:2])

    dir_coeffs[:, order == 1] = (-1j * A / 3) * np.sqrt(4*np.pi) * harm.conj()
    return dir_coeffs
















# ====================== SHD ESTIMATION ======================

def measurement_omni(pos, exp_center, max_order, wave_num):
    T = translation_operator(pos - exp_center, wave_num, max_order, 0)
    T = np.sum(T, axis=-2) # inner product with omni directionality
    return T

def measurement_conj_omni(pos, exp_center, max_order, wave_num):
    """Returns the adjoint measurement operator Xi^H

    Can be applied as shd_coeffs(pos) = Xi^H @ p
    where p is complex sound pressure at the measurement points
    
    Parameters
    ----------
    pos : ndarray of shape (num_pos, 3)
        positions of the measurement points where the sound pressure was measured
    exp_center : ndarray of shape (1, 3)
        expansion center of the spherical harmonics
    max_order = int
        maximum order of the spherical harmonics that is output by the operator
    wave_num : ndarray of shape (num_freqs,)
        wavenumber, defined as w / c where w is the angular frequency
        and c is the speed of sound.
    
    Returns
    -------
    operator : ndarray of shape (num_freqs, num_coeffs, num_pos)
        adjoint of the measurement operator
    """
    T = translation_operator(exp_center - pos, wave_num, 0, max_order)
    T = np.sum(T, axis=-1) # inner product with omni directionality
    return np.moveaxis(T, 1,2)

def apply_measurement_omni(shd_coeffs, pos, exp_center, wave_num):
    """Applies the measurement operator to a sequence of spherical harmonic coefficients
    
    The operator is a function from a complex infinite (l2) sequence to a complex vector
    Xi : l^2(C) -> C^M
    
    Parameters
    ----------
    shd_coeffs : ndarray of shape (num_freq, num_coeffs,)
        spherical harmonic coefficients
    pos : ndarray of shape (num_pos, 3)
        positions of the measurement points where the sound pressure should be estimated
    exp_center : ndarray of shape (1, 3)
        expansion center of the spherical harmonics
    wave_num : ndarray of shape (num_freq,)
        wavenumber, defined as w / c where w is the angular frequency
        and c is the speed of sound.

    Returns
    -------
    p_est : ndarray of shape (num_freq, num_mic,)
        complex sound pressure at the measurement points
    """
    rel_pos = pos - exp_center
    #num_coeffs = shd_coeffs.shape[1]
    #max_order = shd_max_order(num_coeffs) 
    shd_translated = translate_shd_coeffs(shd_coeffs, rel_pos, wave_num, 0)
    p_est = np.sum(shd_translated, axis=-1) # must be multiplied by dir.conj() when not omnidirectional
    return p_est


def apply_measurement_conj_omni(vec, pos, exp_center, max_order, wave_num):
    """Applies the conjugate measurement operator to a complex vector

    The operator is a function from a complex vector to a complex infinite (l2) sequence
    Xi^H : C^M -> l^2(C)
    
    Parameters
    ----------
    vec : ndarray of shape (num_freqs, num_pos)
        complex vector
    pos : ndarray of shape (num_pos, 3)
        positions of the measurement points where the sound pressure was measured
    exp_center : ndarray of shape (1, 3)
        expansion center of the spherical harmonics
    max_order = int
        maximum order of the spherical harmonics that is output by the operator
    wave_num : ndarray of shape (num_freqs,)
        wavenumber, defined as w / c where w is the angular frequency
        and c is the speed of sound. 
    
    Returns
    -------
    shd_coeffs : ndarray of shape (num_freqs, num_coeffs,)
        coefficients of the infinite sequence
    """
    rel_pos = exp_center - pos
    directivity = directivity_omni()
    dir_translated = translate_shd_coeffs(directivity, rel_pos, wave_num, max_order)
    return np.sum(dir_translated * vec[:,:,None], axis=1)

def apply_measurement(shd_coeffs, pos, exp_center, wave_num, dir_coeffs=None):
    """Applies the measurement operator to a sequence of spherical harmonic coefficients
    
    The operator is a function from a complex infinite (l2) sequence to a complex vector
    Xi : l^2(C) -> C^M

    Assumes for now that the same directivity is used for all microphones
    
    Parameters
    ----------
    shd_coeffs : ndarray of shape (num_freq, num_coeffs,)
        spherical harmonic coefficients representing the sound field being measured
    pos : ndarray of shape (num_mic, 3)
        positions of the measurement points where the sound pressure should be estimated
    exp_center : ndarray of shape (1, 3)
        expansion center of the spherical harmonics
    wave_num : ndarray of shape (num_freq,)
        wavenumber, defined as w / c where w is the angular frequency
        and c is the speed of sound.
    dir_coeffs : ndarray of shape (num_freqs, num_coeffs) or (num_freqs, num_mic, num_coeffs), optional
        coefficients of the directivity function. 
        If None, the directivity is assumed to be omnidirectional

    Returns
    -------
    p_est : ndarray of shape (num_freq, num_mic,)
        complex sound pressure at the measurement points
    """
    num_mic = pos.shape[0]
    num_freq = wave_num.shape[0]
    num_coeffs = shd_coeffs.shape[1]
    assert shd_coeffs.shape == (num_freq, num_coeffs,)
    assert pos.shape == (num_mic, 3)
    assert exp_center.shape == (1, 3)
    assert wave_num.shape == (num_freq,)

    rel_pos = pos - exp_center
    if dir_coeffs is None:
        directivity = directivity_omni()
    else:
        directivity = dir_coeffs

    if directivity.ndim == 2:
        directivity = directivity[:,None,:]
    assert directivity.ndim == 3
    assert directivity.shape[0] == num_freq or directivity.shape[0] == 1
    assert directivity.shape[1] == num_mic or directivity.shape[1] == 1

    max_order_dir = shd_max_order(directivity.shape[-1])
    #max_order = shd_max_order(num_coeffs)
    shd_translated = translate_shd_coeffs(shd_coeffs, rel_pos, wave_num, max_order_dir)
    if shd_translated.ndim == 2:
        shd_translated = shd_translated[:,None,:]

    p_est = np.sum(shd_translated * np.conj(directivity), axis=-1)
    return p_est


def apply_measurement_conj(vec, pos, exp_center, max_order, wave_num, dir_coeffs=None):
    """Applies the conjugate measurement operator to a complex vector

    The operator is a function from a complex vector to a complex infinite (l2) sequence
    Xi^H : C^M -> l^2(C)
    
    Parameters
    ----------
    vec : ndarray of shape (num_freqs, num_pos)
        complex vector
    pos : ndarray of shape (num_pos, 3)
        positions of the measurement points where the sound pressure was measured
    exp_center : ndarray of shape (1, 3)
        expansion center of the spherical harmonics
    max_order = int
        maximum order of the spherical harmonics that is output by the operator
    wave_num : ndarray of shape (num_freqs,)
        wavenumber, defined as w / c where w is the angular frequency
        and c is the speed of sound. 
    dir_coeffs : ndarray of shape (num_freqs, num_coeffs) or (num_freqs, num_pos, num_coeffs), optional
        coefficients of the directivity function. If None, the directivity is assumed to be omnidirectional
    
    Returns
    -------
    shd_coeffs : ndarray of shape (num_freqs, num_coeffs,)
        coefficients of the infinite sequence
    """
    rel_pos = exp_center - pos
    if dir_coeffs is None:
        directivity = directivity_omni()
    else:
        directivity = dir_coeffs
    max_order_dir = shd_max_order(directivity.shape[-1])
    
    dir_translated = translate_shd_coeffs(directivity, rel_pos, wave_num, max_order)
    return np.sum(dir_translated * vec[:,:,None], axis=1)

def translated_inner_product(pos1, pos2, dir_coeffs1, dir_coeffs2, wave_num):
    """Computes the inner product of translated sequences of coefficients
    
    $\\langle T(r_1 - r_2, \\omega_m) \\gamma_2(\\omega_m), \\gamma_1(\\omega_m) \\rangle$
    
    Parameters
    ----------
    pos1 : ndarray of shape (num_pos1, 3)
        positions of the first set of measurement points
    pos2 : ndarray of shape (num_pos2, 3)
        positions of the second set of measurement points
    dir_coeffs1 : ndarray of shape (num_freqs, num_coeffs1) or (num_freqs, num_pos1, num_coeffs1)
        coefficients of the directivity function for the first set of measurement points
    dir_coeffs2 : ndarray of shape (num_freqs, num_coeffs2) or (num_freqs, num_pos1, num_coeffs1)
        coefficients of the directivity function for the second set of mea surement points
    wave_num : ndarray of shape (num_freqs,)
        wavenumber, defined as w / c where w is the angular frequency
        and c is the speed of sound.

    Returns
    -------
    psi : ndarray of shape (num_freqs, num_pos1, num_pos2)
        inner product of the translated directivity functions
    """
    assert pos1.ndim == 2 and pos2.ndim == 2
    assert pos1.shape[-1] == 3 and pos2.shape[-1] == 3
    num_pos1 = pos1.shape[0]
    num_pos2 = pos2.shape[0]

    assert wave_num.ndim == 1
    num_freqs = wave_num.shape[0]
    
    if dir_coeffs1.ndim == 2:
        dir_coeffs1 = dir_coeffs1[:,None,:]
    if dir_coeffs2.ndim == 2:
        dir_coeffs2 = dir_coeffs2[:,None,:]
    assert dir_coeffs1.ndim == 3 and dir_coeffs2.ndim == 3
    assert dir_coeffs1.shape[1] == num_pos1 or dir_coeffs1.shape[1] == 1
    assert dir_coeffs2.shape[1] == num_pos2 or dir_coeffs2.shape[1] == 1
    assert dir_coeffs1.shape[0] == num_freqs or dir_coeffs1.shape[0] == 1
    assert dir_coeffs2.shape[0] == num_freqs or dir_coeffs2.shape[0] == 1

    max_order1 = shd_max_order(dir_coeffs1.shape[-1])

    pos_diff = pos1[:,None,:] - pos2[None,:,:]

    translated_coeffs2 = np.stack([translate_shd_coeffs(dir_coeffs2, pos_diff[m,:,:], wave_num, max_order1) for m in range(num_pos1)], axis=1)
    inner_product_matrix = np.sum(translated_coeffs2 * dir_coeffs1.conj()[:,:,None,:], axis=-1)
    return inner_product_matrix

def inf_dimensional_shd(p, pos, exp_center, max_order, wave_num, reg_param, dir_coeffs=None):
    """Estimates the spherical harmonic coefficients with Bayesian inference, allows for arbitrary directivity
    
    Implements the method in Natsuki Ueno, Shoichi Koyama, Hiroshi Saruwatai,
    'Sound Field Recording Using Distributed Microphones Based on Harmonic Analysis of Infinite Order'

    Parameters
    ----------
    p : ndarray of shape (num_real_freqs, num_mics)
        complex sound pressure for each microphone and frequency
    pos : ndarray of shape (num_mics, 3)
        positions of the microphones
    exp_center : ndarray of shape (1, 3)
        expansion center of the spherical harmonics
    max_order : int
        maximum order of the spherical harmonics
    wave_num : float
        wavenumber, defined as w / c where w is the angular frequency
        and c is the speed of sound
    reg_param : float
        regularization parameter. Must be non-negative
    dir_coeffs : ndarray of shape (num_real_freqs, num_coeffs) or (num_real_freqs, num_mics, num_coeffs)
        coefficients of the directivity function. If None, the directivity is assumed to be omnidirectional
        if the array is 2-dimensional, the same directivity is assumed for all microphones
    
    Returns
    -------
    shd_coeffs : complex ndarray of shape (num_real_freqs, num_coeffs)
        harmonic coefficients of the estimated sound field
    """
    assert exp_center.shape == (3,) or exp_center.shape == (1, 3)
    num_mic = pos.shape[0]

    psi = translated_inner_product(pos, pos, dir_coeffs, dir_coeffs, wave_num)
    psi_plus_noise_cov = psi + np.eye(num_mic) * reg_param
    regression_vec = np.linalg.solve(psi_plus_noise_cov,  p)

    shd_coeffs = apply_measurement_conj(regression_vec, pos, exp_center, max_order, wave_num, dir_coeffs=dir_coeffs)
    return shd_coeffs


def inf_dimensional_shd_omni(p, pos, exp_center, max_order, wave_num, reg_param):
    """Estimates the spherical harmonic coefficients with Bayesian inference
    
    Assumes all microphones are omnidirectional.
    Is identical to posterior_mean_omni, if reg_param = noise_power / prior_variance. So 
    this function should be a wrapper for posterior_mean_omni or removed. 

    Parameters
    ----------
    p : ndarray of shape (num_real_freqs, num_mics)
        complex sound pressure for each microphone and frequency
    pos : ndarray of shape (num_mics, 3)
        positions of the microphones
    exp_center : ndarray of shape (1, 3)
        expansion center of the spherical harmonics
    max_order : int
        maximum order of the spherical harmonics
    wave_num : ndarray of shape (num_real_freqs,)
        wavenumber, defined as w / c where w is the angular frequency
        and c is the speed of sound
    reg_param : float or ndarray of shape (num_real_freqs,)
        regularization parameter. Must be non-negative
    
    Returns
    -------
    shd_coeffs : complex ndarray of shape (num_real_freqs, num_coeffs)
        harmonic coefficients of the estimated sound field

    References
    ----------
    Natsuki Ueno, Shoichi Koyama, Hiroshi Saruwatai, 'Sound Field Recording Using Distributed 
    Microphones Based on Harmonic Analysis of Infinite Order'
    """
    num_mic = pos.shape[0]
    num_freq = wave_num.shape[0]

    if isinstance(reg_param, np.ndarray):
        assert reg_param.ndim == 1
        assert reg_param.shape[0] == num_freq
        assert np.all(reg_param >= 0)
    else:
        assert reg_param >= 0
        reg_param = np.ones(num_freq) * reg_param
    noise_cov = np.eye(num_mic)[None,...] * reg_param[:,None,None]

    psi = ki.kernel_helmholtz_3d(pos, pos, wave_num)
    psi_plus_noise_cov = psi + noise_cov
    regression_vec = np.linalg.solve(psi_plus_noise_cov,  p)

    shd_coeffs = apply_measurement_conj_omni(regression_vec, pos, exp_center, max_order, wave_num)
    return shd_coeffs

def inf_dimensional_shd_omni_prior(p, pos, exp_center, max_order, wave_num, reg_param, prior_mean):
    """Estimates spherical harmonic coefficients with Bayesian inference of an infinite sequence of spherical harmonics

    Implements the method in Natsuki Ueno, Shoichi Koyama, Hiroshi Saruwatai,
    'Sound Field Recording Using Distributed Microphones Based on Harmonic Analysis of Infinite Order'

    Parameters
    ----------
    p : ndarray of shape (num_real_freqs, num_mics)
        complex sound pressure for each microphone and frequency
    pos : ndarray of shape (num_mics, 3)
        positions of the microphones
    exp_center : ndarray of shape (1, 3)
        expansion center of the spherical harmonics
    max_order : int
        maximum order of the spherical harmonics
    reg_param : float
        regularization parameter. Must be non-negative
    wave_num : ndarray of shape (num_real_freqs,)
        wavenumber, defined as w / c where w is the angular frequency
        and c is the speed of sound
    prior_mean : ndarray of shape (num_real_freqs, num_coeffs,)
        mean of the prior distribution for the sherical harmonic coefficients
    
    Returns
    -------
    shd_coeffs : complex ndarray of shape (num_real_freqs, num_coeffs)
        harmonic coefficients of the estimated sound field
    """
    num_mic = pos.shape[0]

    psi = ki.kernel_helmholtz_3d(pos, pos, wave_num) 
    psi_plus_noise_cov = psi + np.eye(num_mic) * reg_param
    p_prior = apply_measurement_omni(prior_mean, pos, exp_center, max_order, wave_num)
    regression_vec = np.linalg.solve(psi_plus_noise_cov,  p - p_prior)

    shd_coeffs = prior_mean + apply_measurement_conj_omni(regression_vec, pos, exp_center, max_order, wave_num)
    return shd_coeffs




def posterior_mean_omni(p, pos, exp_center, max_order, wave_num, prior_covariance, noise_power, prior_mean=None):
    """Computes the posterior mean of the spherical harmonic coefficients, under a Gaussian assumption
    
    Parameters
    ----------
    p : ndarray of shape (num_real_freqs, num_mics)
        complex sound pressure for each microphone and frequency
    pos : ndarray of shape (num_mics, 3)
        positions of the microphones
    exp_center : ndarray of shape (1, 3)
        expansion center of the spherical harmonics
    max_order : int
        maximum order of the spherical harmonics
    wave_num : ndarray of shape (num_real_freqs,)
        wavenumber, defined as w / c where w is the angular frequency
        and c is the speed of sound
    noise_power : float
        variance of the noise in the measurement equation. The noise covariance matrix
        is assumed to be a identity matrix scaled by this value
    prior_mean : ndarray of shape (num_real_freqs, num_coeffs), optional
        mean of the prior distribution for the sherical harmonic coefficients
        If not provided, the prior mean is assumed to be zero
    prior_covariance : ndarray of shape (num_real_freqs, num_coeffs, num_coeffs), 
        or ndarray of shape (num_real_freqs,) if the covariance is a scalar matrix
        or float if the covariance is a scalar matrix, identical for all frequencies. 

    Returns
    -------
    shd_coeffs : complex ndarray of shape (num_real_freqs, num_coeffs)
        posterior mean of the harmonic coefficients
    """
    num_mic = pos.shape[0]
    num_freq = wave_num.shape[0]
    num_coeffs = shd_num_coeffs(max_order)
    assert p.shape == (num_freq, num_mic)
    assert wave_num.ndim == 1

    measure = measurement_omni(pos, exp_center, max_order, wave_num)
    measure_conj = measurement_conj_omni(pos, exp_center, max_order, wave_num)

    # Prior mean
    if prior_mean is not None:
       p_prior = measure @ prior_mean[...,None]
       p = p - p_prior[...,0]

    # Prior covariance
    if isinstance(prior_covariance, (int, float)):
        assert prior_covariance >= 0
        prior_covariance = np.ones(num_freq) * prior_covariance
    if prior_covariance.ndim == 1:
        assert prior_covariance.shape[0] == num_freq
        assert np.all(prior_covariance >= 0)
        psi = ki.kernel_helmholtz_3d(pos, pos, wave_num) * prior_covariance[:,None,None]
    elif prior_covariance.ndim == 3:
        assert prior_covariance.shape == (num_freq, num_coeffs, num_coeffs)
        psi = measure @ prior_covariance @ measure_conj
    else:
        raise ValueError("Invalid shape of prior covariance")
  
    # Noise covariance
    if isinstance(noise_power, (int, float)):
        assert noise_power >= 0
        noise_cov = np.eye(num_mic)[None,...] * noise_power
    elif isinstance(noise_power, np.ndarray):
        assert noise_power.ndim == 1
        assert noise_power.shape[0] == num_freq
        assert np.all(noise_power >= 0)
        noise_cov = np.eye(num_mic)[None,...] * noise_power[:,None,None]

    psi_plus_noise_cov = psi + noise_cov
    regression_vec = np.linalg.solve(psi_plus_noise_cov, p)

    if prior_covariance.ndim == 1:
        shd_coeffs = prior_covariance[:,None,None] * measure_conj @ regression_vec[...,None]
    elif prior_covariance.ndim == 3:
        shd_coeffs = prior_covariance @ measure_conj @ regression_vec[...,None]
    shd_coeffs = np.squeeze(shd_coeffs, axis=-1)

    if prior_mean is not None:
        shd_coeffs += prior_mean
    return shd_coeffs

def _posterior_mean_omni_scalar_covariance(p, pos, exp_center, max_order, wave_num, prior_variance, noise_power, prior_mean = None):
    """A special case of posterior_mean_omni_arbitrary_cov, and should be removed eventually
    """
    if prior_mean is not None:
        p_prior = measurement_omni(pos, exp_center, max_order, wave_num) @ prior_mean[...,None]
        p = p - p_prior[...,0]

    num_mic = pos.shape[0]

    psi = ki.kernel_helmholtz_3d(pos, pos, wave_num) 
    psi_plus_noise_cov = psi + np.eye(num_mic) * noise_power / prior_variance 
    regression_vec = np.linalg.solve(psi_plus_noise_cov, p)

    shd_coeffs = measurement_conj_omni(pos, exp_center, max_order, wave_num) @ regression_vec[...,None]#np.linalg.inv(psi_plus_noise_cov) @ p[:,:,None]
    shd_coeffs = np.squeeze(shd_coeffs, axis=-1)

    if prior_mean is not None:
        shd_coeffs += prior_mean
    return shd_coeffs

def _posterior_covariance_omni_scalar_covariance(pos, exp_center, max_order, wave_num, prior_variance, noise_power):
    """ Returns the posterior covariance of the spherical harmonic coefficients

    Assumes Gaussian priors for the spherical harmonic coefficients and Gaussian noise in the measurement equation

    Parameters
    ----------
    pos : ndarray of shape (num_mics, 3)
        positions of the microphones
    exp_center : ndarray of shape (1, 3)
        expansion center of the spherical harmonics
    max_order : int
        maximum order of the spherical harmonics
    wave_num : ndarray of shape (num_real_freqs,)
        wavenumber, defined as w / c where w is the angular frequency
        and c is the speed of sound
    prior_variance : float
        variance of the prior distribution for the sherical harmonic coefficients. The prior
        covariance matrix is assumed to be an identity matrix scaled by this value.
    noise_power : float
        variance of the noise in the measurement equation. The noise covariance matrix
        is assumed to be a identity matrix scaled by this value
    
    Returns
    -------
    cov : ndarray of shape (num_real_freqs, num_coeffs, num_coeffs)
        posterior covariance of the spherical harmonic coefficients
    
    References
    ----------
    Natsuki Ueno, Shoichi Koyama, Hiroshi Saruwatai, 'Sound Field Recording 
    Using Distributed Microphones Based on Harmonic Analysis of Infinite Order'
    """
    num_mic = pos.shape[0]
    num_freq = wave_num.shape[0]
    num_coeffs = shd_num_coeffs(max_order)

    assert prior_variance.ndim == 1
    assert prior_variance.shape[0] == num_freq
    if isinstance(noise_power, (int, float)):
        noise_power = np.ones(num_freq) * noise_power
    assert noise_power.ndim == 1
    assert noise_power.shape[0] == num_freq

        #assert noise_power.shape[0] == num_freq
    #assert isinstance(noise_power, (int, float)) # not implemented for array noise power

    psi = ki.kernel_helmholtz_3d(pos, pos, wave_num) #* prior_variance[:,None,None]
    noise_cov = np.eye(num_mic)[None,:,:] * noise_power[:,None,None] / prior_variance[:,None,None]
    psi_plus_noise_cov = psi + noise_cov

    measure = measurement_omni(pos, exp_center, max_order, wave_num)
    measure_conj = measurement_conj_omni(pos, exp_center, max_order, wave_num)
    cov = measure_conj @ np.linalg.solve(psi_plus_noise_cov, measure)

    # cov = measurement_conj_omni(pos, exp_center, max_order, wave_num) @ \
    #         np.linalg.inv(psi_plus_noise_cov) @ \
    #             measurement_omni(pos, exp_center, max_order, wave_num)
    cov = prior_variance[:,None,None] * (np.eye(num_coeffs)[None,:,:] - cov)
    return cov

def posterior_covariance_omni(pos, exp_center, max_order, wave_num, prior_covariance, noise_power):
    """Returns the posterior covariance of the spherical harmonic coefficients
    Assumes Gaussian priors for the spherical harmonic coefficients and Gaussian noise in the measurement equation
    
    Parameters
    ----------
    pos : ndarray of shape (num_mics, 3)
        positions of the microphones
    exp_center : ndarray of shape (1, 3)
        expansion center of the spherical harmonics
    max_order : int
        maximum order of the spherical harmonics
    wave_num : ndarray of shape (num_real_freqs,)
        wavenumber, defined as w / c where w is the angular frequency
        and c is the speed of sound
    noise_power : float
        variance of the noise in the measurement equation. The noise covariance matrix
        is assumed to be a identity matrix scaled by this value
    prior_covariance : ndarray of shape (num_real_freqs, num_coeffs, num_coeffs), 
        or ndarray of shape (num_real_freqs,) if the covariance is a scalar matrix
        or float if the covariance is a scalar matrix, identical for all frequencies. 
    """
    num_mic = pos.shape[0]
    num_freq = wave_num.shape[0]
    num_coeffs = shd_num_coeffs(max_order)
    assert wave_num.ndim == 1

    if isinstance(prior_covariance, (int, float)):
        assert prior_covariance >= 0
        prior_covariance = np.ones(num_freq) * prior_covariance
    if prior_covariance.ndim == 1:
        return _posterior_covariance_omni_scalar_covariance(pos, exp_center, max_order, wave_num, prior_covariance, noise_power)
    assert prior_covariance.shape == (num_freq, num_coeffs, num_coeffs)

    measure = measurement_omni(pos, exp_center, max_order, wave_num)
    measure_conj = measurement_conj_omni(pos, exp_center, max_order, wave_num)

    # Noise covariance
    if isinstance(noise_power, (int, float)):
        assert noise_power >= 0
        noise_cov = np.eye(num_mic)[None,...] * noise_power
    elif isinstance(noise_power, np.ndarray):
        assert noise_power.ndim == 1
        assert noise_power.shape[0] == num_freq
        assert np.all(noise_power >= 0)
        noise_cov = np.eye(num_mic)[None,...] * noise_power[:,None,None]

    psi = measure @ prior_covariance @ measure_conj
    psi_plus_noise_cov = psi + noise_cov

    measure_times_cov = measure @ prior_covariance
    cov_times_measure_conj = prior_covariance @ measure_conj
    main_mat = cov_times_measure_conj @ np.linalg.solve(psi_plus_noise_cov, measure_times_cov)

    return prior_covariance - main_mat
